package edu.mit.harold.icollections;

import java.util.AbstractCollection;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map.Entry;


/**
 * 
 * A map-backed immutable bag.
 * 
 * @author harold
 *
 * @param <E>
 */
public final class MapIBag<E> extends AbstractCollection<E> implements IBag<E> {
//// STATIC FACTORY METHODS ////
	/**
	 * @param <E>
	 * @param map
	 * @return an IBag backed by an empty version of map, i.e. by map.minusAll(map.keySet())
	 */
	public static <E> MapIBag<E> empty(final IMap<E,Integer> map) {
		return new MapIBag<E>(map.minusAll(map.keySet()), 0, 0); }
	

//// PRIVATE CONSTRUCTORS ////
	private final IMap<E,Integer> map;
	private final int size, hashCode;
	// not instantiable (or subclassable):
	private MapIBag(final IMap<E,Integer> map, final int size, final int hashCode) {
		this.map = map; this.size = size; this.hashCode = hashCode; }


//// REQUIRED METHODS FROM AbstractCollection ////
	@Override
	public int size() {
		return size; }
	@Override
	public Iterator<E> iterator() {
		final Iterator<Entry<E,Integer>> i = map.entrySet().iterator();
		return new Iterator<E>() {
			private E e; private int n=0;
			public boolean hasNext() {
				return n>0 || i.hasNext(); }
			public E next() {
				if(n==0) { // finished with current element
					Entry<E,Integer> entry = i.next();
					e = entry.getKey();
					n = entry.getValue();
				}
				n--;
				return e;
			}
			public void remove() {
				throw new UnsupportedOperationException(); }
		};
	}
	

//// OVERRIDDEN METHODS OF AbstractCollection ////
	@Override
	public boolean contains(final Object e) {
		return map.containsKey(e); }
	@Override
	public int hashCode() {
		return hashCode; }
	
	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Object that) {
		if(!(that instanceof IBag))
			return false;
		if(!(that instanceof MapIBag)) {
			// make that into a MapIBag:
			// TODO this is INEFFICIENT
			MapIBag<Object> empty = (MapIBag<Object>)this.minusAll(this);
			that = empty.plusAll((IBag<?>)that);
		}
		return this.map.equals( ((MapIBag<?>)that).map );
	}

	
//// IMPLEMENTED METHODS OF ISet ////
	public MapIBag<E> plus(final E e) {
		return new MapIBag<E>(map.plus(e, count(e)+1), size+1, hashCode+e.hashCode());
	}
	
	@SuppressWarnings("unchecked")
	public MapIBag<E> minus(final Object e) {
		int n = count(e);
		if(n==0)
			return this;
		if(n==1) // remove from map
			return new MapIBag<E>(map.minus(e), size-1, hashCode-e.hashCode());
		// otherwise just decrement count:
		return new MapIBag<E>(map.plus((E)e, n-1), size-1, hashCode-e.hashCode());
	}

	public MapIBag<E> plusAll(final Collection<? extends E> list) {
		MapIBag<E> bag = this;
		for(E e : list) bag = bag.plus(e);
		return bag;
	}
	
	public MapIBag<E> minusAll(final Collection<?> list) {
		// removes _all_ elements found in list, i.e. counts are irrelevant:
		IMap<E,Integer> map = this.map.minusAll(list);
		return new MapIBag<E>(map, size(map), hashCode(map)); // (completely recomputes size and hashCode)
	}

	
//// PRIVATE UTILITIES ////
	// TODO should this be part of IBag?
	@SuppressWarnings("unchecked")
	private int count(final Object o) {
		if(!contains(o)) return 0;
		// otherwise o must be an E:
		return map.get((E)o);
	}
	
//// PRIVATE STATIC UTILITIES ////
	private static int size(final IMap<?,Integer> map) {
		int size = 0;
		for(Integer n : map.values())
			size += n;
		return size;
	}

	private static int hashCode(final IMap<?,Integer> map) {
		int hashCode = 0;
		for(Entry<?,Integer> e : map.entrySet())
			hashCode += e.getValue() * e.getKey().hashCode();
		return hashCode;
	}
}
