package edu.mit.harold.icollections;

import java.util.AbstractList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map.Entry;



/**
 * 
 * An immutable vector of non-null elements.
 * <p>
 * This implementation is backed by an IntTreeIMap and
 * supports logarithmic-time querying, setting, insertion,
 * and removal.
 * 
 * @author harold
 *
 * @param <E>
 */
public class TreeIVector<E> extends AbstractList<E> implements IVector<E> {
//// STATIC FACTORY METHODS ////
	private static final TreeIVector<Object> EMPTY = new TreeIVector<Object>(IntTreeIMap.empty());
	
	/**
	 * @param <E>
	 * @return an empty vector
	 */
	@SuppressWarnings("unchecked")
	public static <E> TreeIVector<E> empty() {
		return (TreeIVector<E>)EMPTY; }
	
	/**
	 * @param <E>
	 * @param e
	 * @return empty().plus(e)
	 */
	public static <E> TreeIVector<E> singleton(final E e) {
		return TreeIVector.<E>empty().plus(e); }
	
	/**
	 * @param <E>
	 * @param list
	 * @return empty().plusAll(list)
	 */
	@SuppressWarnings("unchecked")
	public static <E> TreeIVector<E> from(final Collection<? extends E> list) {
		if(list instanceof TreeIVector)
			return (TreeIVector<E>)list; //(actually we only know it's TreeIVector<? extends E>)
									// but that's good enough for an immutable
									// (i.e. we can't mess someone else up by adding the wrong type to it)
		return TreeIVector.<E>empty().plusAll(list);
	}

//// PRIVATE CONSTRUCTORS ////
	private final IntTreeIMap<E> map;
	private TreeIVector(final IntTreeIMap<E> map) {
		this.map = map; }
	

//// REQUIRED METHODS FROM AbstractList ////
	@Override
	public int size() {
		return map.size(); }
	
	@Override
	public E get(final int index) {
		if(index<0 || index>=size())
			throw new IndexOutOfBoundsException();
		return map.get(index);
	}
	

//// OVERRIDDEN METHODS FROM AbstractList ////
	private Integer hashCode=null;
	@Override
	public int hashCode() {
		if(hashCode==null)
			hashCode = super.hashCode(); // AbstractList.hashCode(), uses iterator()
		return hashCode;
	}
	
	@Override
	public Iterator<E> iterator() {
		return map.values().iterator(); }

	@Override
	public TreeIVector<E> subList(final int start, final int end) {
		final int size = size();
		if(start<0 || end>size || start>end)
			throw new IndexOutOfBoundsException();
		if(start==end)
			return empty();
		if(start==0) {
			if(end==size) return this;
			// remove from end:
			return this.minus(size-1).subList(start, end);
		}
		// remove from start:
		return this.minus(0).subList(start-1,end-1);
	}
	

////IMPLEMENTED METHODS OF IVector ////
	public TreeIVector<E> plus(final E e) {
		return new TreeIVector<E>( map.plus(size(), e) ); }

	public TreeIVector<E> plus(final int i, final E e) {
		if(i<0 || i>size())
			throw new IndexOutOfBoundsException();
		return new TreeIVector<E>( map.withKeysChangedAbove(i,1).plus(i,e) );
	}

	public TreeIVector<E> minus(final Object e) {
		for(Entry<Integer,E> entry : map.entrySet())
			if(entry.getValue().equals(e))
				return minus((int)entry.getKey());
		return this;
	}

	public TreeIVector<E> minus(final int i) {
		if(i<0 || i>=size())
			throw new IndexOutOfBoundsException();
		return new TreeIVector<E>( map.minus(i).withKeysChangedAbove(i,-1) );
	}

	public TreeIVector<E> plusAll(final Collection<? extends E> list) {
		TreeIVector<E> result = this;
		for(E e : list)
			result = result.plus(e);
		return result;
	}

	public TreeIVector<E> minusAll(final Collection<?> list) {
		TreeIVector<E> result = this;
		for(Object e : list)
			result = result.minus(e);
		return result;
	}

	public TreeIVector<E> plusAll(int i, final Collection<? extends E> list) {
		if(list.size()==0)
			return this;
		IntTreeIMap<E> map = this.map.withKeysChangedAbove(i, list.size());
		for(E e : list)
			map = map.plus(i++, e);
		return new TreeIVector<E>( map );
	}

	public IVector<E> with(final int i, final E e) {
		// TODO add OutOfBoundsExceptions ??
		// or maybe get rid of them everywhere; they're not in my spec...
		IntTreeIMap<E> map = this.map.plus(i, e);
		if(map==this.map) return this;
		return new TreeIVector<E>( map );
	}
}
